Index: amd64/identcpu.c
===================================================================
RCS file: /cvs/src/sys/arch/amd64/amd64/identcpu.c,v
retrieving revision 1.32
diff -u -r1.32 identcpu.c
--- amd64/identcpu.c	29 May 2011 14:50:25 -0000	1.32
+++ amd64/identcpu.c	10 Jul 2011 16:11:15 -0000
@@ -52,6 +52,7 @@
 #ifdef CRYPTO
 int amd64_has_aesni;
 #endif
+int have_mwait;
 
 const struct {
 	u_int32_t	bit;
@@ -368,12 +369,14 @@
 			setperf_setup = k1x_init;
 	}
 
-	if (cpu_ecxfeature & CPUIDECX_EST) {
+	if (cpu_ecxfeature & CPUIDECX_EST)
 		setperf_setup = est_init;
-	}
 
 	if (cpu_ecxfeature & CPUIDECX_AES)
 		amd64_has_aesni = 1;
+
+	if (cpu_ecxfeature & CPUIDECX_MWAIT)
+		have_mwait = 1;
 
 	if (!strncmp(cpu_model, "Intel", 5)) {
 		u_int32_t cflushsz;
Index: amd64/machdep.c
===================================================================
RCS file: /cvs/src/sys/arch/amd64/amd64/machdep.c,v
retrieving revision 1.148
diff -u -r1.148 machdep.c
--- amd64/machdep.c	5 Jul 2011 04:48:01 -0000	1.148
+++ amd64/machdep.c	10 Jul 2011 16:11:16 -0000
@@ -147,7 +147,12 @@
  * switchto vectors
  */
 void (*cpu_idle_leave_fcn)(void) = NULL;
+#ifdef MULTIPROCESSOR
+void cpu_idle_mwait(void);
+void (*cpu_idle_cycle_fcn)(void) = cpu_idle_mwait;
+#else
 void (*cpu_idle_cycle_fcn)(void) = NULL;
+#endif
 void (*cpu_idle_enter_fcn)(void) = NULL;
 
 /* the following is used externally for concurrent handlers */
@@ -186,6 +191,8 @@
 int kbd_reset;
 int lid_suspend;
 
+extern int have_mwait;
+
 /*
  * safepri is a safe priority for sleep to set for a spin-wait
  * during autoconfiguration or after a panic.
@@ -695,11 +702,60 @@
 }
 
 #ifdef MULTIPROCESSOR
+
+#define MWAIT_WAKEUP 0x1
+#define MWAIT_WAITING 0x2
+/*
+ * mwait cpu power states. Lower 4 bits are sub-states.
+ */
+#define MWAIT_C0 0xf0
+#define MWAIT_C1 0x00
+#define MWAIT_C2 0x10
+#define MWAIT_C3 0x20
+#define MWAIT_C4 0x30
+
+static __inline void
+cpu_monitor(const void *addr, int extensions, int hints)
+{
+	__asm __volatile("monitor;"
+		: :"a" (addr), "c" (extensions), "d"(hints));
+}
+
+static __inline void
+cpu_mwait(int extensions, int hints)
+{  
+	__asm __volatile("mwait;" : :"a" (hints), "c" (extensions));
+}
+
+void
+cpu_idle_mwait()
+{
+	int *mwait;
+
+	if (!have_mwait) {
+		__asm __volatile("sti;hlt");
+		return;
+	}
+	mwait = (int*)curcpu()->ci_mmwbuf;
+	*mwait = MWAIT_WAITING;
+	cpu_monitor(mwait, 0, 0);
+	if (*mwait == MWAIT_WAITING)
+		cpu_mwait(0, MWAIT_C1);
+}
+
 void
 cpu_unidle(struct cpu_info *ci)
 {
-	if (ci != curcpu())
-		x86_send_ipi(ci, X86_IPI_NOP);
+	int *mwait;
+
+	if (ci != curcpu()) {
+		if (have_mwait) {
+			mwait = (int*)ci->ci_mmwbuf;
+			*mwait = MWAIT_WAKEUP;
+		} else {
+			x86_send_ipi(ci, X86_IPI_NOP);
+		}
+	}
 }
 #endif
 
Index: include/cpu.h
===================================================================
RCS file: /cvs/src/sys/arch/amd64/include/cpu.h,v
retrieving revision 1.68
diff -u -r1.68 cpu.h
--- include/cpu.h	16 Jun 2011 19:46:40 -0000	1.68
+++ include/cpu.h	10 Jul 2011 16:11:16 -0000
@@ -128,6 +128,8 @@
 
 	struct ksensordev	ci_sensordev;
 	struct ksensor		ci_sensor;
+
+	char	ci_mmwbuf[128] __attribute__ ((aligned(128)));
 };
 
 #define CPUF_BSP	0x0001		/* CPU is the original BSP */
